from collections import defaultdict
import numpy as np
from pyquaternion import Quaternion
from typing import List

from google.protobuf.json_format import MessageToDict
from .lidar_frame_1_pb2 import CameraImage, LidarFrame

from .transform import Transform


# Protobuf Helpers

def create_scene_from_protobufs(scene, protobufs: List[str]):
    """
    Create a LidarScene object from a list of protobuf files
    Args:
        scene: LidarScene object
        protobufs: List of filepaths to the protobuf files

    Returns: The updated LidarScene

    """
    for frame_num, protobuf in enumerate(protobufs):
        with open(protobuf, "rb") as f:
            frame = LidarFrame.FromString(f.read())

        pose = Transform.from_Rt(
            R=Quaternion(
                frame.device_heading.w,
                frame.device_heading.x,
                frame.device_heading.y,
                frame.device_heading.z,
            ),
            t=[
                frame.device_position.x,
                frame.device_position.y,
                frame.device_position.z,
            ],
        )

        # Group points by device ID
        points_by_sensor = defaultdict(list)
        for point in frame.points:
            sensor_id = getattr(point, "d", 0)
            points_by_sensor[sensor_id].append(point)

        for sensor_id, lidar_points in points_by_sensor.items():
            points = np.asarray([[p.x, p.y, p.z, p.i] for p in lidar_points])
            scene.get_frame(frame_num).add_points(points, sensor_id=sensor_id)

        scene.get_frame(frame_num).apply_transform(pose)

        for camera in frame.images:
            camera_num = camera.camera_index
            image_url = camera.image_url

            # Calibrate cameras once
            if frame_num == 0:
                calibrate_camera(scene, camera)

            scene.get_frame(frame_num).get_image(camera_num).load_file(image_url)

    return scene


def calibrate_camera(scene, camera: CameraImage):
    camera_pose = Transform.from_Rt(
        R=Quaternion(
            camera.heading.w,
            camera.heading.x,
            camera.heading.y,
            camera.heading.z,
        ),
        t=[camera.position.x, camera.position.y, camera.position.z],
    )

    distortion_model = camera.WhichOneof("camera_intrinsics")
    intrinsics = getattr(camera, distortion_model)

    # Protobuf supports xi for omnidirectional but not supported by calibrate
    if hasattr(intrinsics, "xi"):
        print('NOTE: For omnnidirectional intrinsics, xi is not supported.')

    intrinsics_params = MessageToDict(intrinsics)

    scene.get_camera(camera.camera_index).calibrate(
        pose=camera_pose, **intrinsics_params
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: lidar_frame_1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x13lidar_frame_1.proto"\\\n\x05Point\x12\t\n\x01x\x18\x01 \x02(\x02\x12\t\n\x01y\x18\x02 \x02(\x02\x12\t\n\x01z\x18\x03 \x02(\x02\x12\t\n\x01i\x18\x04 \x01(\x02\x12\t\n\x01\x64\x18\x05 \x01(\r\x12\t\n\x01t\x18\x06 \x01(\x01\x12\x11\n\tis_ground\x18\x07 \x01(\x08"*\n\x07Vector3\x12\t\n\x01x\x18\x01 \x02(\x02\x12\t\n\x01y\x18\x02 \x02(\x02\x12\t\n\x01z\x18\x03 \x02(\x02"8\n\nQuaternion\x12\t\n\x01w\x18\x01 \x02(\x02\x12\t\n\x01x\x18\x02 \x02(\x02\x12\t\n\x01y\x18\x03 \x02(\x02\x12\t\n\x01z\x18\x04 \x02(\x02"5\n\x07GPSPose\x12\x0b\n\x03lat\x18\x01 \x02(\x02\x12\x0c\n\x04long\x18\x02 \x02(\x02\x12\x0f\n\x07\x62\x65\x61ring\x18\x03 \x02(\x02"S\n\nRadarPoint\x12\x1a\n\x08position\x18\x01 \x02(\x0b\x32\x08.Vector3\x12\x1b\n\tdirection\x18\x02 \x01(\x0b\x32\x08.Vector3\x12\x0c\n\x04size\x18\x03 \x01(\x02"\x9e\x01\n\x16\x42rownConradyIntrinsics\x12\n\n\x02\x66x\x18\x01 \x02(\x02\x12\n\n\x02\x66y\x18\x02 \x02(\x02\x12\n\n\x02\x63x\x18\x03 \x02(\x02\x12\n\n\x02\x63y\x18\x04 \x02(\x02\x12\x0c\n\x04skew\x18\x05 \x01(\x02\x12\n\n\x02k1\x18\x06 \x01(\x02\x12\n\n\x02k2\x18\x07 \x01(\x02\x12\n\n\x02p1\x18\x08 \x01(\x02\x12\n\n\x02p2\x18\t \x01(\x02\x12\n\n\x02k3\x18\n \x01(\x02J\x04\x08\x0b\x10\x0cJ\x04\x08\x0c\x10\r"\x81\x01\n\x11\x46isheyeIntrinsics\x12\n\n\x02\x66x\x18\x01 \x02(\x02\x12\n\n\x02\x66y\x18\x02 \x02(\x02\x12\n\n\x02\x63x\x18\x03 \x02(\x02\x12\n\n\x02\x63y\x18\x04 \x02(\x02\x12\x0c\n\x04skew\x18\x05 \x01(\x02\x12\n\n\x02k1\x18\x06 \x01(\x02\x12\n\n\x02k2\x18\x07 \x01(\x02\x12\n\n\x02k3\x18\x08 \x01(\x02\x12\n\n\x02k4\x18\t \x01(\x02"\xa1\x01\n\x19OmnidirectionalIntrinsics\x12\n\n\x02\x66x\x18\x01 \x02(\x02\x12\n\n\x02\x66y\x18\x02 \x02(\x02\x12\n\n\x02\x63x\x18\x03 \x02(\x02\x12\n\n\x02\x63y\x18\x04 \x02(\x02\x12\x0c\n\x04skew\x18\x05 \x01(\x02\x12\n\n\x02k1\x18\x06 \x01(\x02\x12\n\n\x02k2\x18\x07 \x01(\x02\x12\n\n\x02p1\x18\x08 \x01(\x02\x12\n\n\x02p2\x18\t \x01(\x02\x12\n\n\x02k3\x18\n \x01(\x02\x12\n\n\x02xi\x18\x0b \x01(\x02"\xe2\x02\n\x0b\x43\x61meraImage\x12\x11\n\ttimestamp\x18\x01 \x01(\x01\x12\x11\n\timage_url\x18\x02 \x02(\t\x12\x1a\n\x08position\x18\x03 \x02(\x0b\x32\x08.Vector3\x12\x1c\n\x07heading\x18\x04 \x02(\x0b\x32\x0b.Quaternion\x12\x14\n\x0cscale_factor\x18\x05 \x01(\x02\x12\x10\n\x08priority\x18\x06 \x01(\x05\x12\x14\n\x0c\x63\x61mera_index\x18\x07 \x01(\x05\x12\x30\n\rbrown_conrady\x18\x08 \x01(\x0b\x32\x17.BrownConradyIntrinsicsH\x00\x12%\n\x07\x66isheye\x18\t \x01(\x0b\x32\x12.FisheyeIntrinsicsH\x00\x12\x35\n\x0fomnidirectional\x18\r \x01(\x0b\x32\x1a.OmnidirectionalIntrinsicsH\x00\x42\x13\n\x11\x63\x61mera_intrinsicsJ\x04\x08\n\x10\x0bJ\x04\x08\x0b\x10\x0cJ\x04\x08\x0c\x10\r"\xf2\x01\n\nLidarFrame\x12!\n\x0f\x64\x65vice_position\x18\x01 \x02(\x0b\x32\x08.Vector3\x12#\n\x0e\x64\x65vice_heading\x18\x02 \x01(\x0b\x32\x0b.Quaternion\x12!\n\x0f\x64\x65vice_gps_pose\x18\x03 \x01(\x0b\x32\x08.GPSPose\x12\x16\n\x06points\x18\x04 \x03(\x0b\x32\x06.Point\x12!\n\x0cradar_points\x18\x05 \x03(\x0b\x32\x0b.RadarPoint\x12\x1c\n\x06images\x18\x06 \x03(\x0b\x32\x0c.CameraImage\x12\x11\n\ttimestamp\x18\x07 \x01(\x01\x12\r\n\x05\x65xtra\x18\x08 \x01(\x0c'
)


_POINT = DESCRIPTOR.message_types_by_name["Point"]
_VECTOR3 = DESCRIPTOR.message_types_by_name["Vector3"]
_QUATERNION = DESCRIPTOR.message_types_by_name["Quaternion"]
_GPSPOSE = DESCRIPTOR.message_types_by_name["GPSPose"]
_RADARPOINT = DESCRIPTOR.message_types_by_name["RadarPoint"]
_BROWNCONRADYINTRINSICS = DESCRIPTOR.message_types_by_name["BrownConradyIntrinsics"]
_FISHEYEINTRINSICS = DESCRIPTOR.message_types_by_name["FisheyeIntrinsics"]
_OMNIDIRECTIONALINTRINSICS = DESCRIPTOR.message_types_by_name[
    "OmnidirectionalIntrinsics"
]
_CAMERAIMAGE = DESCRIPTOR.message_types_by_name["CameraImage"]
_LIDARFRAME = DESCRIPTOR.message_types_by_name["LidarFrame"]
Point = _reflection.GeneratedProtocolMessageType(
    "Point",
    (_message.Message,),
    {
        "DESCRIPTOR": _POINT,
        "__module__": "lidar_frame_1_pb2"
        # @@protoc_insertion_point(class_scope:Point)
    },
)
_sym_db.RegisterMessage(Point)

Vector3 = _reflection.GeneratedProtocolMessageType(
    "Vector3",
    (_message.Message,),
    {
        "DESCRIPTOR": _VECTOR3,
        "__module__": "lidar_frame_1_pb2"
        # @@protoc_insertion_point(class_scope:Vector3)
    },
)
_sym_db.RegisterMessage(Vector3)

Quaternion = _reflection.GeneratedProtocolMessageType(
    "Quaternion",
    (_message.Message,),
    {
        "DESCRIPTOR": _QUATERNION,
        "__module__": "lidar_frame_1_pb2"
        # @@protoc_insertion_point(class_scope:Quaternion)
    },
)
_sym_db.RegisterMessage(Quaternion)

GPSPose = _reflection.GeneratedProtocolMessageType(
    "GPSPose",
    (_message.Message,),
    {
        "DESCRIPTOR": _GPSPOSE,
        "__module__": "lidar_frame_1_pb2"
        # @@protoc_insertion_point(class_scope:GPSPose)
    },
)
_sym_db.RegisterMessage(GPSPose)

RadarPoint = _reflection.GeneratedProtocolMessageType(
    "RadarPoint",
    (_message.Message,),
    {
        "DESCRIPTOR": _RADARPOINT,
        "__module__": "lidar_frame_1_pb2"
        # @@protoc_insertion_point(class_scope:RadarPoint)
    },
)
_sym_db.RegisterMessage(RadarPoint)

BrownConradyIntrinsics = _reflection.GeneratedProtocolMessageType(
    "BrownConradyIntrinsics",
    (_message.Message,),
    {
        "DESCRIPTOR": _BROWNCONRADYINTRINSICS,
        "__module__": "lidar_frame_1_pb2"
        # @@protoc_insertion_point(class_scope:BrownConradyIntrinsics)
    },
)
_sym_db.RegisterMessage(BrownConradyIntrinsics)

FisheyeIntrinsics = _reflection.GeneratedProtocolMessageType(
    "FisheyeIntrinsics",
    (_message.Message,),
    {
        "DESCRIPTOR": _FISHEYEINTRINSICS,
        "__module__": "lidar_frame_1_pb2"
        # @@protoc_insertion_point(class_scope:FisheyeIntrinsics)
    },
)
_sym_db.RegisterMessage(FisheyeIntrinsics)

OmnidirectionalIntrinsics = _reflection.GeneratedProtocolMessageType(
    "OmnidirectionalIntrinsics",
    (_message.Message,),
    {
        "DESCRIPTOR": _OMNIDIRECTIONALINTRINSICS,
        "__module__": "lidar_frame_1_pb2"
        # @@protoc_insertion_point(class_scope:OmnidirectionalIntrinsics)
    },
)
_sym_db.RegisterMessage(OmnidirectionalIntrinsics)

CameraImage = _reflection.GeneratedProtocolMessageType(
    "CameraImage",
    (_message.Message,),
    {
        "DESCRIPTOR": _CAMERAIMAGE,
        "__module__": "lidar_frame_1_pb2"
        # @@protoc_insertion_point(class_scope:CameraImage)
    },
)
_sym_db.RegisterMessage(CameraImage)

LidarFrame = _reflection.GeneratedProtocolMessageType(
    "LidarFrame",
    (_message.Message,),
    {
        "DESCRIPTOR": _LIDARFRAME,
        "__module__": "lidar_frame_1_pb2"
        # @@protoc_insertion_point(class_scope:LidarFrame)
    },
)
_sym_db.RegisterMessage(LidarFrame)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _POINT._serialized_start = 23
    _POINT._serialized_end = 115
    _VECTOR3._serialized_start = 117
    _VECTOR3._serialized_end = 159
    _QUATERNION._serialized_start = 161
    _QUATERNION._serialized_end = 217
    _GPSPOSE._serialized_start = 219
    _GPSPOSE._serialized_end = 272
    _RADARPOINT._serialized_start = 274
    _RADARPOINT._serialized_end = 357
    _BROWNCONRADYINTRINSICS._serialized_start = 360
    _BROWNCONRADYINTRINSICS._serialized_end = 518
    _FISHEYEINTRINSICS._serialized_start = 521
    _FISHEYEINTRINSICS._serialized_end = 650
    _OMNIDIRECTIONALINTRINSICS._serialized_start = 653
    _OMNIDIRECTIONALINTRINSICS._serialized_end = 814
    _CAMERAIMAGE._serialized_start = 817
    _CAMERAIMAGE._serialized_end = 1171
    _LIDARFRAME._serialized_start = 1174
    _LIDARFRAME._serialized_end = 1416
# @@protoc_insertion_point(module_scope)
